
/**
 *
 *  This code is part of the publication:
 *
 *      "Contours, corners and T-junctions detection algorithm" by Antoni Buades,
 *      Rafael Grompone von Gioi, and Julia Navarro. Image Processing On Line, 2018.
 *
 *  Antoni Buades <toni.buades@uib.es> Universitat de les Illes Balears, Spain
 *  Rafael Grompone von Gioi <grompone@gmail.com> CMLA, ENS Cachan, France
 *  Julia Navarro <julia.navarro@uib.es> Universitat de les Illes Balears, Spain
 *
 *
 **/

/**
 *  Implementation of the contour detection method proposed in
 *
 *          Antonio Buades, Rafael Grompone von Gioi, and Julia Navarro, "Joint contour,
 *          corner and t-junction detection: An approach inspired by the mammal visual
 *          system", Journal of Mathematical Imaging and Vision, (2017), pp. 1-14.
 *
 *
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include <vector>
#include <sstream>


#include "libImage.h"
#include "libBasic.h"
#include "libDetectionIpol.h"


#define MAX(i,j) ( (i)<(j) ? (j):(i) )
#define MIN(i,j) ( (i)<(j) ? (i):(j) )

using namespace std;

// This function generates the output files.
void save(float sigma, std::vector < std::vector <str_point_descriptor> > &llista, vector<str_point_descriptor> &cCorners,char *fileCorners, char * imOutputGray, char * imOutputColor,  char *fileTextContours, int width, int height)
{

    libUSTG::cflimage output(width, height, 3); output=255.0f;

    //! Curves image
    float *outputRed = output.v(0);
    float *outputGreen = output.v(1);
    float *outputBlue = output.v(2);


    //! grayscale output
    libUSTG::flimage gray(width,height);
    gray = 255.0f;


    std::ofstream oconfile;
    oconfile.open(fileTextContours);


    //! Fill data
    int iaSize = 0;
    srand(0);
    for (int kk=0; kk < (int) llista.size(); kk++)
    {

        float red =  (float) drand48();
        float green =  (float) drand48();
        float blue =   (float) drand48();


        vector<str_point_descriptor> corba;
        corba = llista[kk];

        for (int ii=0; ii < (int) corba.size(); ii++)
        {

            outputRed[corba[ii].py   * width + corba[ii].px] = 255.0f * red;
            outputGreen[corba[ii].py * width + corba[ii].px] = 255.0f *  green;
            outputBlue[corba[ii].py  * width + corba[ii].px] = 255.0f *  blue;

            gray[corba[ii].py * width + corba[ii].px] = 0.0f;


            oconfile <<  corba[ii].px <<  " " << corba[ii].py <<  " " ;

            iaSize++;

        }


        oconfile << std::endl;

    }


    oconfile.close();

    std::ofstream ocfile;
    ocfile.open(fileCorners);

    for (int kk=0; kk < (int) cCorners.size(); kk++)
    {

        int px = cCorners[kk].px;
        int py = cCorners[kk].py;
        float color = 128.0f;

        if (cCorners[kk].ncorbes>2)
            color = 0.0f;

        ocfile << px << " " << py << " " << cCorners[kk].ncorbes <<  " "  << std::endl;


        libUSTG::fiImageDrawCircle(outputRed, px, py, 5.0f, color, width, height);
        libUSTG::fiImageDrawCircle(outputGreen, px, py, 5.0f, color, width, height);
        libUSTG::fiImageDrawCircle(outputBlue, px, py, 5.0f, color, width, height);

        libUSTG::fiImageDrawCircle(gray.v(), px, py, 5.0f, color, width, height);

    }

    ocfile.close();


    output.save(imOutputColor);
    gray.save(imOutputGray);

}





// Detection algorithm.
int main(int argc, char **argv)
{


    vector <OptStruct *> options;

    OptStruct oS = {"s:", 0,   "2.0",  NULL, "sigma"};  options.push_back(&oS);

    OptStruct oT = {"t:", 0,   "2.5",  NULL, "threshold for lateral inhibition"};  options.push_back(&oT);

    vector<ParStruct *> parameters;
    ParStruct pinput = {"input", NULL, "input image"}; parameters.push_back(&pinput);
    ParStruct pfile = {"file", NULL, "input probability file"}; parameters.push_back(&pfile);
    ParStruct pout = {"out", NULL, "output edges"}; parameters.push_back(&pout);
    ParStruct poutc = {"out color", NULL, "output color edges"}; parameters.push_back(&poutc);
    ParStruct confile = {"file", NULL, "output text file containing contour"}; parameters.push_back(&confile);
    ParStruct corfile = {"file", NULL, "output text file containing corners"}; parameters.push_back(&corfile);


    if (!parsecmdline("src_detection_derivative_gaussians", "filter image by directional gaussian derivative", argc, argv, options, parameters))
        return EXIT_FAILURE;




    //! Parameters
    float fSigma = atof(oS.value);
    float anglePrecision = 7.5f;
    int connectivity = 2;          //! Connectivity for grouping       default: 2 (5x5x5)

    int nDir;
    nDir = (int) rintf(360 / anglePrecision);


    //! First Inhibition
    float fThresholdLI = atof(oT.value);        //! Threshold for first inhibitions default: 2.5
    float fAngleFiltPrec = 15.0f;               //! degres used for inter-orientation inhibitions and filtering

    //! Directional Convolution
    //!  float sigma  = Sigma;                  //! Parameters of oriented filter
    //!  float sigmay = 1.0f;

    //! Corner identification
    float fCorThreshold = 0.275;
    fCorThreshold = fCorThreshold * (float) nDir;
    float minRespCorner = 0.005;
    float radForZone = fSigma;
    float fCorValidationDegree = 15.0;


    //! Corner clearing
    float radForClearCor = 2.0;          //! 3 * sigma, radius of circular zone to search for arriving contours


    //! Second Inhibition
    float fThresholdLISecondInhibitions = 0.005;


    //! Parameters: Input Images
    libUSTG::cflimage inputImageC;
    inputImageC.load(pinput.value);
    libUSTG::flimage inputImage = inputImageC.getGray();
    


    int width = inputImage.w();
    int height = inputImage.h();
    int wh = width*height;
    float *input = inputImage.v();


    //!
    //! Compute directional derivatives
    //!
    float **dimages = new float*[nDir];
    for (int i=0; i < nDir; i++)  dimages[i] = new float[width*height];

    printf("Compute directional convolutions ...\n");
    {
        compute_directional_convolutions(input, dimages,  nDir,  fSigma,  anglePrecision, width, height);

    }




    //!
    //! Lateral inhibition
    //!
    printf("Lateral inhibition ...\n");
    {
        float **dFilt = new float*[nDir];
        for (int i=0; i < nDir; i++)  dFilt[i] = new float[wh];

        int flagKeepValue = 0;
        lateral_inhibition_each_scale(dimages, dFilt, fSigma, nDir, width, height, fAngleFiltPrec, fThresholdLI, anglePrecision, flagKeepValue);

        for (int i=0; i < nDir; i++)  libUSTG::fpCopy(dFilt[i], dimages[i], width*height);

        //! Delete memory
        for (int i=0; i < nDir; i++) delete[] dFilt[i];
        delete[]  dFilt;
    }



    //!
    //! Good continuation filtering
    //!
    printf("Good continuation filter ...\n");
    {
        float **dFilt = new float*[nDir];
        for (int i=0; i < nDir; i++)  dFilt[i] = new float[wh];


        good_continuation_filter(dimages, dFilt, fSigma, fAngleFiltPrec, nDir, anglePrecision, width, height);

        for (int i=0; i < nDir; i++)  libUSTG::fpCopy(dFilt[i], dimages[i], width*height);

        //!
        //! Delete memory
        //!
        for (int i=0; i < nDir; i++) delete[] dFilt[i];
        delete[]  dFilt;
    }



    //!
    //! Compute local maxima of transversal average as corner indicator
    //!
    float *localMax=new float[width*height];
    libUSTG::fpClear(localMax, 255.0, width*height);
    vector<str_point_descriptor> cCorners;
    printf("Compute corners ...\n");
    {
        compute_corners(dimages, localMax, cCorners, nDir, fSigma, minRespCorner, fCorThreshold, radForZone, width, height);
    }



    //!
    //! Second Lateral inhibition
    //!
    {
        printf("Second lateral inhibition ...\n");
        float **dFilt = new float*[nDir];
        for (int i=0; i < nDir; i++)  dFilt[i] = new float[wh];

        int flagKeepValue = 1;              //! 1: keep original non inhibited values.  0: set to 1 non inhibited values
        lateral_inhibition_each_scale(dimages, dFilt, fSigma, nDir, width, height, fAngleFiltPrec, fThresholdLISecondInhibitions, anglePrecision, flagKeepValue);

        for (int i=0; i < nDir; i++)  libUSTG::fpCopy(dFilt[i], dimages[i], width*height);

        //!
        //! Delete memory
        //!
        for (int i=0; i < nDir; i++) delete[] dFilt[i];
        delete[]  dFilt;
    }




    //!
    //! Perform first grouping
    //!
    printf("Grouping gestalt binomial ...\n");
    float gestMult = 4.0f;
    float ntestsM = 1.0 / (gestMult * fSigma * fSigma);

    float lnfaThr = 0.0f;

    std::vector < std::vector <str_point_descriptor> > llista;
    float **curvesID = new float*[nDir];
    for (int i=0; i < nDir; i++)  curvesID[i] = new float[wh];
    {

        float fThreshold = 0.0f;  //! Take into account all points with positive density
        grouping_gestalt_binomial(dimages,curvesID, localMax,fThreshold, llista,  width, height, nDir, pfile.value, ntestsM, lnfaThr, connectivity);

    }





    //!
    //! Clean corners: erase corners not joining curves of different orientation.
    //! dir and dir+pi are taken as the same orientation in order to remove spurious corners on stroke lines
    //! curve grouping is redone, and t-junctions detected afterwards
    //!
    {
        //! clear corners
        clear_corners(cCorners, localMax, curvesID, fSigma, nDir, anglePrecision, radForClearCor, radForZone, fCorValidationDegree, width,  height);

        llista.clear();
        
        //! recompute curves
        float fThreshold = 0.0f;
        grouping_gestalt_binomial(dimages,curvesID, localMax,fThreshold, llista,  width, height, nDir, pfile.value, ntestsM, lnfaThr, connectivity );
    }


    //!
    //! Differentiate corners from T-junctions
    //!
    {
        differentiate_corners_junctions(cCorners, curvesID, fSigma, nDir, radForClearCor, width,  height);
    }


    //!
    //! Save results
    //!
    char *fileCorners=NULL;
    fileCorners=corfile.value;

    char *fileTextContours= confile.value;

    save(fSigma, llista, cCorners, fileCorners, pout.value, poutc.value, fileTextContours, width, height);

    for (int i=0; i < nDir; i++) delete[] dimages[i];
    delete[] dimages;


    return EXIT_SUCCESS;

}

